@@ -87,7 +87,8 @@ func (r *PerformanceProfileReconciler) SetupWithManager(mgr ctrl.Manager) error
 			kubeletOld := e.ObjectOld.(*mcov1.KubeletConfig)
 			kubeletNew := e.ObjectNew.(*mcov1.KubeletConfig)
 
-			return !reflect.DeepEqual(kubeletOld.Status.Conditions, kubeletNew.Status.Conditions)
+			return kubeletOld.GetGeneration() != kubeletNew.GetGeneration() ||
+				!reflect.DeepEqual(kubeletOld.Status.Conditions, kubeletNew.Status.Conditions)
 		},
 	}
 
@@ -223,6 +224,7 @@ func validateUpdateEvent(e *event.UpdateEvent) bool {
 // +kubebuilder:rbac:groups=machineconfiguration.openshift.io,resources=machineconfigs;machineconfigpools;kubeletconfigs,verbs=*
 // +kubebuilder:rbac:groups=tuned.openshift.io,resources=tuneds;profiles,verbs=*
 // +kubebuilder:rbac:groups=node.k8s.io,resources=runtimeclasses,verbs=*
+// +kubebuilder:rbac:groups=config.openshift.io,resources=infrastructures,verbs=get;list;watch
 // +kubebuilder:rbac:namespace="openshift-performance-addon-operator",groups=core,resources=pods;services;services/finalizers;configmaps,verbs=*
 // +kubebuilder:rbac:namespace="openshift-performance-addon-operator",groups=coordination.k8s.io,resources=leases,verbs=create;get;list;update
 // +kubebuilder:rbac:namespace="openshift-performance-addon-operator",groups=apps,resourceNames=performance-operator,resources=deployments/finalizers,verbs=update
